/*
Title:
	C source code file for EM4095 RFID reader IC

Datasheet:
	http://www.emmicroelectronic.com/webfiles/Product/RFID/DS/EM4095_DS.pdf

Application note:
	http://www.emmicroelectronic.com/webfiles/product/rfid/an/an404.pdf

Compiler:
	avrgcc 3.4.5 with AVR Studio 4.12 Service Pack 3 Build 490

Author:
	Stefan Seegel / dahamm@gmx.net

Date:
	30.08.2006

Version:
	1.0

Free for noncommercial use
*/


#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h>
#include "em4100.h"

#define MANCHESTER_LONG (F_CPU / F_RFID * 64 / TIMER0_PRESCALER)
#define MANCHESTER_SHORT (F_CPU / F_RFID * 64 / TIMER0_PRESCALER / 2)
#define MANCHESTER_MIDDLE (MANCHESTER_LONG + MANCHESTER_SHORT) / 2

static unsigned char data[8];
static volatile uint8_t flag;

ISR(INT1_vect)
{
	static uint16_t sync;
	static uint8_t bitcount;
	static uint8_t insync = 0;
	static uint8_t old_cnt;
	if ((uint8_t)(TCNT0 - old_cnt) > MANCHESTER_MIDDLE)
	{
		old_cnt = TCNT0;

		sync <<= 1;
		bitcount++;
		
		if (bit_is_clear(PIND, PD3))
			sync |= 1;
		
		if ((sync & 0x3FF) == 0x1FF)
		{
			data[0] = 0xFF;
			bitcount = 1;
			insync = 1;
		}

		if (insync)
		{
			if ((bitcount % 8) == 0)
			{
				data[bitcount / 8] = sync & 0xFF;

				if (bitcount == 56)
				{
					flag = 1;
					insync = 0;
					GIMSK &= ~(1<<INT1);
				}
			}
		}
	}
}

void rfid_init(void)
{
	DDRD |= (1<<DDD2);
}

void rfid_enable(void)
{
	PORTD |= (1<<PORTD2);
}

void rfid_disable(void)
{
	PORTD &= ~(1<<PORTD2);
}

static uint8_t check_parity(uint16_t param)
{
	uint8_t bit;
	uint8_t par = 0;
	
	for (bit=0; bit<16; bit++)
		par ^= (param >> bit) & 0x01;

	if (par)
		return 0;
	else
		return 1;
}


uint8_t em4095_read_tag(uint8_t *d)
{
	uint8_t timeout = 50;

	do
	{
		if (bit_is_set(TIFR, TOV0))
		{
			timeout--;
			TIFR |= 1<<TOV0; //Clear Overflow Flag
		}

	} while ((!flag) && (timeout > 0));
	

	if (!flag)
		return 0;


	flag = 0;
	uint8_t temp[8];
	memcpy(temp, data, 8);
	memset(data, 0, 8);
	GIMSK |= 1<<INT1;

	if (((temp[0] << 8 | temp[1]) & 0xFF80) != 0xFF80) //Check startbits
		return 0;
	
	if ((temp[7] & 0x01)) //Check stopbit
		return 0;
	
	if (!check_parity(temp[1] & 0x7C))
		return 0;
	
	if (!check_parity((temp[1] & 0x03) | (temp[2] & 0xE0)))
		return 0;
	
	if (!check_parity(temp[2] & 0x1F))
		return 0;
			
	if (!check_parity(temp[3] & 0xF8))
		return 0;
		
	if (!check_parity((temp[3] & 0x07) | (temp[4] & 0xC0)))
		return 0;
			
	if (!check_parity(temp[4] & 0x3E))
		return 0;
			  
	if (!check_parity((temp[4] & 0x01) | (temp[5] & 0xF0)))
		return 0;
			
	if (!check_parity((temp[5] & 0x0F) | (temp[6] & 0x80)))
		return 0;
				  
	if (!check_parity(temp[6] & 0x7C))
		return 0;
			  
	if (!check_parity((temp[6] & 0x03) | (temp[7] & 0xE0)))
		return 0;
		
	if (!check_parity((temp[1] & 0x40) ^ (temp[1] & 0x02) ^ (temp[2] & 0x10) ^ (temp[3] & 0x80) ^ (temp[3] & 0x04) ^
									  (temp[4] & 0x20) ^ (temp[4] & 0x01) ^ (temp[5] & 0x08) ^ (temp[6] & 0x40) ^ (temp[6] & 0x02) ^
									  (temp[7] & 0x10)))
		return 0;
		
	if (!check_parity((temp[1] & 0x20) ^ (temp[1] & 0x01) ^ (temp[2] & 0x08) ^ (temp[3] & 0x40) ^ (temp[3] & 0x02) ^
									  (temp[4] & 0x10) ^ (temp[5] & 0x80) ^ (temp[5] & 0x04) ^ (temp[6] & 0x20) ^ (temp[6] & 0x01) ^
									  (temp[7] & 0x08)))
		return 0;
	
	if (!check_parity((temp[1] & 0x10) ^ (temp[2] & 0x80) ^ (temp[2] & 0x04) ^ (temp[3] & 0x20) ^ (temp[3] & 0x01) ^
									  (temp[4] & 0x08) ^ (temp[5] & 0x40) ^ (temp[5] & 0x02) ^ (temp[6] & 0x10) ^ (temp[7] & 0x80) ^
									  (temp[7] & 0x04)))
		return 0;
	
	if (!check_parity((temp[1] & 0x08) ^ (temp[2] & 0x40) ^ (temp[2] & 0x02) ^ (temp[3] & 0x10) ^ (temp[4] & 0x80) ^
									  (temp[4] & 0x04) ^ (temp[5] & 0x20) ^ (temp[5] & 0x01) ^ (temp[6] & 0x08) ^ (temp[7] & 0x40) ^
									  (temp[7] & 0x02)))
		return 0;

//11111111 11000110 00101111 10001100 01010011 00010000 00100101 00110110
//          DDDD DD DD DDDD  DDDD DDD D DDDD D DDD DDDD  DDDD DD DD DDDD
//          0000 00 00 1111  1111 222 2 2222 3 333 3333  4444 44 44


	d[0] = (temp[1] & 0x78) << 1 | (temp[1] & 0x03) << 2 | (temp[2] & 0xC0) >> 6;
	d[1] = (temp[2] & 0x1E) << 3 | (temp[3] & 0xF0) >> 4;
	d[2] = (temp[3] & 0x07) << 5 | (temp[4] & 0x80) >> 3 | (temp[4] & 0x3C) >> 2;
	d[3] = (temp[4] & 0x01) << 7 | (temp[5] & 0xE0) >> 1 | (temp[5] & 0x0F);
	d[4] = (temp[6] & 0x78) << 1 | (temp[6] & 0x03) << 2 | (temp[7] & 0xC0) >> 6;


	return 1;
}
